#  Copyright (c) 2019.  Carsten Blank
# 
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
# 
#       http://www.apache.org/licenses/LICENSE-2.0
# 
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from marshmallow.fields import String, Integer, List, DateTime, Boolean
from marshmallow.validate import Range, Length, OneOf, Equal, Regexp
from qiskit.validation import BaseModel, BaseSchema, bind_schema


class AcQuantumGateConfigSchema(BaseSchema):
    """Schema for AcQuantumGateConfig."""

    # Required properties.
    name = String(required=True)
    parameters = List(String(), required=True)
    qasm_def = String(required=True)

    # Optional properties.
    coupling_map = List(List(Integer(),
                             validate=Length(min=1)),
                        validate=Length(min=1))
    latency_map = List(List(Integer(validate=OneOf([0, 1])),
                            validate=Length(min=1)),
                       validate=Length(min=1))
    conditional = Boolean()
    description = String()


class AcQuantumBackendConfigurationSchema(BaseSchema):
    """Schema for AcQuantumBackendConfiguration."""

    # Required properties.
    backend_name = String(required=True)
    backend_version = String(required=True,
                             validate=Regexp("[0-9]+.[0-9]+.[0-9]+$"))
    n_qubits = Integer(required=True, validate=Range(min=1))
    basis_gates = List(String(), required=True,
                       validate=Length(min=1))
    # gates = Nested(AcQuantumGateConfigSchema, required=True, many=True,
    #                validate=Length(min=1))
    local = Boolean(required=True)
    simulator = Boolean(required=True)
    conditional = Boolean(required=True)
    open_pulse = Boolean(required=True, validate=Equal(False))
    memory = Boolean(required=True)
    max_shots = Integer(required=True, validate=Range(min=1))

    # Optional properties.
    max_experiments = Integer(validate=Range(min=1))
    sample_name = String()
    coupling_map = List(List(Integer(),
                             validate=Length(min=1)),
                        validate=Length(min=1))
    n_registers = Integer(validate=Range(min=1))
    register_map = List(List(Integer(validate=OneOf([0, 1])),
                             validate=Length(min=1)),
                        validate=Length(min=1))
    configurable = Boolean()
    credits_required = Boolean()
    online_date = DateTime()
    display_name = String()
    description = String()
    tags = List(String())


@bind_schema(AcQuantumBackendConfigurationSchema)
class AcQuantumBackendConfiguration(BaseModel):
    """Model for BackendConfiguration.

    Please note that this class only describes the required fields. For the
    full description of the model, please check ``AcQuantumBackendConfigurationSchema``.
    Attributes:
        backend_name (str): backend name.
        backend_version (str): backend version in the form X.Y.Z.
        n_qubits (int): number of qubits.
        basis_gates (list[str]): list of basis gates names on the backend.
        gates (GateConfig): list of basis gates on the backend.
        local (bool): backend is local or remote.
        simulator (bool): backend is a simulator.
        conditional (bool): backend supports conditional operations.
        open_pulse (bool): backend supports open pulse.
        memory (bool): backend supports memory.
        max_shots (int): maximum number of shots supported.
    """

    def __init__(self, backend_name, backend_version, n_qubits, basis_gates, gates, local, simulator, conditional,
                 open_pulse, memory, max_shots, **kwargs):
        # type: (str, str, int, [str], 'AcQuantumGateConfig', bool, bool , bool, bool, bool, int, dict) -> None

        super().__init__(**kwargs)
        self.backend_name = backend_name
        self.backend_version = backend_version
        self.n_qubits = n_qubits
        self.basis_gates = basis_gates
        self.gates = gates
        self.local = local
        self.simulator = simulator
        self.conditional = conditional
        self.open_pulse = open_pulse
        self.memory = memory
        self.max_shots = max_shots


@bind_schema(AcQuantumGateConfigSchema)
class AcQuantumGateConfig(BaseModel):
    """Model for AcQuantumGateConfig.

    Please note that this class only describes the required fields. For the
    full description of the model, please check ``AcQuantumGateConfigSchema``.

    Attributes:
        name (str): the gate name as it will be referred to in QASM.
        parameters (list[str]): variable names for the gate parameters (if any).
        qasm_def (str): definition of this gate in terms of QASM primitives U
            and CX.
    """

    def __init__(self, name, parameters, qasm_def, **kwargs):
        # type: (str, [str], str, dict) -> None
        self.name = name
        self.parameters = parameters
        self.qasm_def = qasm_def

        super().__init__(**kwargs)



#  Copyright (c) 2019.  Carsten Blank
# 
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
# 
#       http://www.apache.org/licenses/LICENSE-2.0
# 
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from qiskit.providers import BaseBackend
from qiskit.qobj import Qobj

from acquantumconnector.connector.acquantumconnector import AcQuantumConnector
from acquantumconnector.credentials.credentials import AcQuantumCredentials
from acquantumconnector.model.backendtype import AcQuantumBackendType
from acquantumconnector.model.errors import AcQuantumRequestError
from .acquantumerrors import AcQuantumError, AcQuantumBackendError
from .acquantumjob import AcQuantumJob
from .backendconfiguration import AcQuantumBackendConfiguration


class AcQuantumBackend(BaseBackend):

    def __init__(self, configuration, provider, credentials, api):
        # type: (AcQuantumBackendConfiguration, 'AcQuantumProvider', AcQuantumCredentials, AcQuantumConnector) -> None
        """
        :param configuration: configuration of backend
        :param provider:
        :param credentials:
        :param api: api for communicating with Alibaba Computing Quantum
        """
        super().__init__(provider=provider, configuration=configuration)

        self._api = api
        self._credentials = credentials
        try:
            self._backend_type = AcQuantumBackendType[configuration.backend_name]
        except KeyError:
            raise AcQuantumError('Unknown Backend Name')

    def run(self, qobj, job_name=None):
        # type: (Qobj, str) -> AcQuantumJob
        job = AcQuantumJob(self, None, self._api, not self.configuration().simulator, qobj=qobj, job_name=job_name)
        job.submit()
        return job

    def properties(self):
        # TODO: Implement backend properties
        pass

    def status(self):
        if self._is_device():
            return self._api.get_backend_config().system_status

        raise AcQuantumBackendError('Could not find status for Simulator')

    def jobs(self, limit=50, skip=0):
        # type: (int, int) -> [AcQuantumJob]
        """

        :param limit: number of jobs to retrieve
        :param skip: starting index of retrieval
        :return: list of AcQuantumJob instances
        :raises: AcQuantumRequestError
        """

        job_info_list = self._api.get_experiments()

        jobs = []

        for job in job_info_list:
            if job.experiment_type is self._backend_type:
                jobs.append(AcQuantumJob(self, job.experiment_id, self._api, self._is_device(), job_name=job.name))

        if skip:
            jobs = jobs[skip:]

        return jobs[:limit]

    def retrieve_job(self, job_id):
        # type: (str) -> AcQuantumJob
        """
        :param job_id: job id of the job to retrieve
        :return: job: AcQuantum Job
        :raises: AcQuantumBackendError: if retrieval fails
        """
        try:
            response = self._api.get_experiment(int(job_id))
        except AcQuantumRequestError as ex:
            raise AcQuantumBackendError('Failed to get job "{}" {}'.format(job_id, str(ex)))

        return AcQuantumJob(self, response.detail.experiment_id, self._api, self._is_device(),
                            job_name=response.detail.name)

    def _is_device(self):
        # type: () -> bool
        return not bool(self.configuration().simulator)

    def backend_type(self):
        return self._backend_type



#  Copyright (c) 2019.  Carsten Blank
# 
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
# 
#       http://www.apache.org/licenses/LICENSE-2.0
# 
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from collections import OrderedDict, _OrderedDictItemsView

from acquantumconnector.connector.acquantumconnector import AcQuantumConnector
from acquantumconnector.credentials.credentials import AcQuantumCredentials
from jsonschema import ValidationError
from qiskit.providers import BaseProvider

from .acquantumbackend import AcQuantumBackend
from .backendconfiguration import AcQuantumBackendConfiguration


class AcQuantumSingleProvider(BaseProvider):

    def __init__(self, credentials, provider):
        # type: (AcQuantumCredentials, 'AcQuantumProvider') -> None
        super().__init__()

        self._ac_provider = provider
        self.credentials = credentials  # type: AcQuantumCredentials
        self._api = self._authenticate(self.credentials)  # type: AcQuantumConnector

        self._backends = self._discover_remote_backends()

    def backends(self, name=None, **kwargs):
        # type: (str, dict) -> _OrderedDictItemsView
        """
        :param name:  name of backends
        :param kwargs:
        :return:
        """
        backends = self._backends.items()
        if name:
            backends = [b for n, b in backends if n == name]

        return backends

    def _discover_remote_backends(self):
        # type: () -> OrderedDict[str, AcQuantumBackend]

        ret = OrderedDict()
        configs_list = self._api.available_backends()

        for raw_config in configs_list:
            try:
                config = AcQuantumBackendConfiguration.from_dict(raw_config)
                ret[config.backend_name] = AcQuantumBackend(
                    configuration=config,
                    provider=self._ac_provider,
                    credentials=self.credentials,
                    api=self._api)
            except ValidationError as ex:
                print(
                    'Remote backend {} could not be instantiated due to an invalid config: {}'.format(
                        raw_config.get('backend_name', raw_config.get('name', 'unknown')), ex)
                )

        return ret

    @classmethod
    def _authenticate(cls, credentials):
        # type: (AcQuantumCredentials) -> AcQuantumConnector
        connector = AcQuantumConnector()
        connector.create_session(credentials)
        return connector



#  Copyright (c) 2019.  Carsten Blank
# 
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
# 
#       http://www.apache.org/licenses/LICENSE-2.0
# 
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import json
from enum import Enum
from typing import Any


class AcQuantumBackendType(Enum):
    REAL = 'REAL',
    SIMULATE = 'SIMULATE'


class AcQuantumRequestError(Exception):
    def __init__(self, message, status_code=None):
        # type: (str, int) -> None
        self.status_code = status_code
        self.message = message

    def __str__(self):
        return 'AcRequestError: \'{}\''.format(self.message)

    def __repr__(self):
        return 'AcRequestError: \'{}\''.format(self.message)


class AcQuantumRequestForbiddenError(AcQuantumRequestError):
    def __init__(self, message='403 Forbidden'):
        # type: (str) -> None
        self.message = message

    def __str__(self):
        return '403 - AcRequestForbiddenError: \'{}\''.format(self.message)

    def __repr__(self):
        return '403 - AcRequestForbiddenError: \'{}\''.format(self.message)


class AcQuantumResponse:
    def __init__(self, success=True, data=None, exception=None):
        # type: (bool,Any, Any) -> None
        self.success = success
        self.data = data
        self.exception = exception

    def __str__(self):
        return 'AcResponse: {{ success: {}, exception: {} }}'.format(self.success, self.exception)

    def __repr__(self):
        return 'AcResponse: {{ success: {}, exception: {} }}'.format(self.success, self.exception)


class AcQuantumErrorResponse(AcQuantumResponse):

    def __init__(self, success=False, exception=None, status_code=None):
        # type: (bool, Any, int) -> None
        super(AcQuantumErrorResponse, self).__init__()
        self.status_code = status_code
        self.success = success
        self.exception = exception

    def __str__(self):
        return 'AcErrorResponse: {{ success: {}, exception: {}, status_code: {} }}'.format(self.success, self.exception,
                                                                                           self.status_code)

    def __repr__(self):
        return 'AcErrorResponse: {{ success: {}, exception: {}, status_code: {} }}'.format(self.success, self.exception,
                                                                                           self.status_code)


class AcQuantumExperimentDetail:

    def __init__(self, name, version, experiment_id, experiment_type, execution, bit_width=None):
        # type: (str, int, int, AcQuantumBackendType, int, int) -> None
        self.name = name
        self.version = version
        self.experiment_id = experiment_id
        self.experiment_type = experiment_type
        self.execution = execution
        self.bit_width = bit_width

    def __str__(self):
        return 'AcExperimentDetail: {{ name: {}, version: {}, experiment_id: {}, experiment_type: {}, execution: {}, ' \
               'bit_width {} }}' \
            .format(self.name, self.version, self.experiment_id, self.experiment_type, self.execution, self.bit_width)

    def __repr__(self):
        return 'AcExperimentDetail: {{ name: {}, version: {}, experiment_id: {}, experiment_type: {}, execution: {}, ' \
               'bit_width {} }}' \
            .format(self.name, self.version, self.experiment_id, self.experiment_type, self.execution, self.bit_width)


class AcQuantumExperiment:

    def __init__(self, detail, data, code=''):
        # type: (AcQuantumExperimentDetail, Any, str) -> None
        self.detail = detail
        self.data = data
        self.code = code

    def __str__(self):
        return 'AcExperiment: {{ data: {}, code: {}, detail: {} }}'.format(self.data, self.code, self.detail)

    def __repr__(self):
        return 'AcExperiment: {{ data: {}, code: {}, detail: {} }}'.format(self.data, self.code, self.detail)


class AcQuantumRawConfig:

    def __init__(self, system_config, one_q_gate_fidelities, qubit_parameter, system_status, two_q_gate_fidelity):
        self.system_config = BackendSystemConfig.from_dict(system_config)
        self.one_q_gate_fidelities = OneQGateFidelities.from_dict(one_q_gate_fidelities)
        self.qubit_parameter = QubitParameter.from_dict(qubit_parameter)
        self.system_status = SystemStatus.from_dict(system_status)
        self.two_q_gate_fidelity = TwoQGateFidelity.from_dict(two_q_gate_fidelity)

    @classmethod
    def from_json(cls, values):
        # type: ([dict]) -> AcQuantumRawConfig
        return AcQuantumRawConfig(*values)


class BackendSystemConfig:

    def __init__(self, computer_id, config_key, config_value):
        # type: (str, str, SystemConfigValue) -> None
        self.config_value = config_value  # type: SystemConfigValue
        self.config_key = config_key
        self.computer_id = computer_id

    @classmethod
    def from_dict(cls, values):
        # type: (dict) -> BackendSystemConfig
        config_value = SystemConfigValue.from_dict(json.loads(values['configValue']))
        return BackendSystemConfig(values['computerId'], values['configKey'], config_value)


class SystemConfigValue:

    def __init__(self, one_q_gates, one_q_gates_label, two_q_gates, two_q_gates_label, measure_size_upper_limit):
        # type: ([str], [str], [str], [str], int) -> None
        self.one_q_gates = one_q_gates
        self.one_q_gates_label = one_q_gates_label
        self.two_q_gates = two_q_gates
        self.two_q_gates_label = two_q_gates_label
        self.measure_size_upper_limit = measure_size_upper_limit

    @classmethod
    def from_dict(cls, values):
        # type: (dict) -> SystemConfigValue
        return SystemConfigValue(values['oneQGates'], values['oneQGatesLabel'], values['twoQGates'],
                                 values['twoQGatesLabel'], values['measureSizeUpperLimit'])


class OneQGateFidelities:

    def __init__(self, computer_id, config_key, config_value):
        self.config_value = config_value  # [dict]
        self.config_key = config_key
        self.computer_id = computer_id

    @classmethod
    def from_dict(cls, values):
        # type: (dict) -> OneQGateFidelities
        config_value = json.loads(values['configValue'])
        return OneQGateFidelities(values['computerId'], values['configKey'], config_value)


class QubitParameter:

    def __init__(self, computer_id, config_key, config_value):
        self.config_value = config_value
        self.config_key = config_key
        self.computer_id = computer_id

    @classmethod
    def from_dict(cls, values):
        # type: (dict) -> QubitParameter
        return QubitParameter(values['computerId'], values['configKey'], values['configValue'])


class SystemStatus:

    def __init__(self, computer_id, config_key, config_value):
        # type: (str, str, SystemStatusConfigValue) -> None
        self.config_value = config_value  # type: SystemStatusConfigValue
        self.config_key = config_key
        self.computer_id = computer_id

    @classmethod
    def from_dict(cls, values):
        # type: (dict) -> SystemStatus
        config_value = SystemStatusConfigValue.from_dict(json.loads(values['configValue']))
        return SystemStatus(values['computerId'], values['configKey'], config_value)


class SystemStatusConfigValue:

    def __init__(self, status, fridge_temperature, last_calibration_time):
        # type: (str, str, str) -> None
        self.last_calibration_time = last_calibration_time
        self.fridge_temperature = fridge_temperature
        self.status = status

    @classmethod
    def from_dict(cls, values):
        # type: (dict) -> SystemStatusConfigValue
        return SystemStatusConfigValue(values['status'], values['fridgeTemperature'], values['lastCalibrationTime'])


class TwoQGateFidelity:

    def __init__(self, computer_id, config_key, config_value):
        self.config_value = config_value  # type: [dict]
        self.config_key = config_key
        self.computer_id = computer_id

    @classmethod
    def from_dict(cls, values):
        # type: (dict) -> TwoQGateFidelity
        config_value = json.loads(values['configValue'])
        return TwoQGateFidelity(values['computerId'], values['configKey'], config_value)



#  Copyright (c) 2019.  Carsten Blank
# 
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
# 
#       http://www.apache.org/licenses/LICENSE-2.0
# 
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import os

from acquantumconnector.credentials.credentials import AcQuantumCredentials
from providers.acquantum.acquantumerrors import AcQuantumAccountError

VAR_MAP = {
    'ACQ_USER': 'user_name',
    'ACQ_PWD': 'password'
}


def read_credentials_from_environ() -> AcQuantumCredentials or None:
    if not (os.getenv('ACQ_USER') and os.getenv('ACQ_PWD')):
        raise AcQuantumAccountError('No Credentials Found in Environment')

    credentials = {}
    for envar_name, credentials_key in VAR_MAP.items():
        if os.getenv(envar_name):
            credentials[credentials_key] = os.getenv(envar_name)

    credentials = AcQuantumCredentials(**credentials)
    return credentials



#  Copyright (c) 2019.  Carsten Blank
# 
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
# 
#       http://www.apache.org/licenses/LICENSE-2.0
# 
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from acquantumconnector.credentials.credentials import AcQuantumCredentials
from providers.acquantum.acquantumerrors import AcQuantumAccountError
from providers.acquantum.credentials._env import read_credentials_from_environ


def discover_credentials() -> AcQuantumCredentials:
    credentials = None
    readers = {
        'environment variables': (read_credentials_from_environ, {}),
    }

    for display_name, (reader_function, kwargs) in readers.items():
        try:
            credentials = reader_function(**kwargs)
            if credentials:
                break
        except AcQuantumAccountError as ex:
            print('Automatic discovery of {} credentials failed: {}'.format(display_name, str(ex)))

    if credentials is None:
        raise AcQuantumAccountError('No Credentials Found')
    else:
        return credentials



#  Copyright (c) 2019.  Carsten Blank
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from qiskit.qiskiterror import QISKitError


class AcQuantumError(QISKitError):
    def __init__(self, *message):
        """Set the error message."""
        super().__init__(' '.join(message))
        self.message = ' '.join(message)

    def __str__(self):
        """Return the message."""
        return repr(self.message)


class AcQuantumAccountError(AcQuantumError):
    """Base class for errors raised by account management."""

    def __init__(self, *message):
        super().__init__(*message)
        self.message = ' '.join(message)

    def __str__(self):
        """Return the message."""
        return repr(self.message)


class AcQuantumJobError(AcQuantumError):
    """Base class for errors raised by job errors"""

    def __init__(self, *message):
        super().__init__(*message)
        self.message = ' '.join(message)

    def __str__(self):
        """Return the message."""
        return repr(self.message)


class AcQuantumJobTimeOutError(AcQuantumJobError):

    def __init__(self, *message):
        super().__init__(*message)
        self.message = ' '.join(message)

    def __str__(self):
        """Return the message."""
        return repr(self.message)


class AcQuantumBackendError(AcQuantumError):

    def __init__(self, *message):
        super().__init__(*message)
        self.message = ' '.join(message)

    def __str__(self):
        """Return the message."""
        return repr(self.message)



#  Copyright (c) 2019.  Carsten Blank
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import datetime
import json
import math
import time
from enum import Enum
from typing import Any, List, Dict, Union

import qiskit
from qiskit import QuantumCircuit, QuantumRegister
from qiskit.dagcircuit import DAGCircuit
from qiskit.providers import BaseJob, BaseBackend
from qiskit.providers.models import BackendConfiguration
from qiskit.qobj import Qobj
from qiskit.result import Result

from acquantumconnector.model.errors import AcQuantumRequestError
from acquantumconnector.model.gates import Gate
from acquantumconnector.model.response import AcQuantumResultResponse, AcQuantumResult
from acquantumconnector.connector.acquantumconnector import AcQuantumConnector
from .acquantumerrors import AcQuantumJobError
from .acquantumerrors import AcQuantumJobTimeOutError
from .models import AcQuantumExperiment


class AcQuantumJobStatus(Enum):
    """Class for job status enumerated type."""
    INITIALIZING = 'job is being initialized'
    QUEUED = 'job is queued'
    VALIDATING = 'job is being validated'
    RUNNING = 'job is actively running'
    CANCELLED = 'job has been cancelled'
    DONE = 'job has successfully run'
    ERROR = 'job incurred error'


JOB_FINAL_STATES = (
    AcQuantumJobStatus.DONE,
    AcQuantumJobStatus.CANCELLED,
    AcQuantumJobStatus.ERROR
)


class AcQuantumJob(BaseJob):
    """
        Represent the jobs that will be executed on Alibaba Computing Quantum simulators and real
        devices. Jobs are intended to be created calling ``run()`` on a particular
        backend.

        Creating a ``AcQuantumJob`` instance does not imply running it. You need to do it in separate steps::

            job = AcQuantumJob(..)
            job.submit() # will block!
    """

    def __init__(self, backend, job_id, api, is_device, qobj=None, creation_date=None, api_status=None, job_name=None):
        # type: (AcQuantumBackend, str, AcQuantumConnector, bool, Qobj, Any, AcQuantumJobStatus, str) -> None
        """
        :param backend: The backend instance used to run this job
        :param job_id: The job ID of an already submitted job
        :param api: api for communicating with Alibaba Computing Quantum
        :param is_device: whether backend is a real device
        :param qobj: Quantum Object
        :param creation_date:
        :param api_status:


        """

        super().__init__(backend, job_id)

        if qobj is not None:
            # validate_qobj_against_schema(qobj)
            self._qobj = qobj

        self._api = api  # type: AcQuantumConnector
        self._backend = backend  # type: 'AcQuantumBackend'
        self._cancelled = False
        self._status = AcQuantumJobStatus.INITIALIZING
        self._job_name = job_name
        # In case of not providing a `qobj`, it is assumed the job already
        # exists in the API (with `job_id`).

        if qobj is None:
            # Some API calls (`get_status_jobs`, `get_status_job`) provide
            # enough information to recreate the `Job`. If that is the case, try
            # to make use of that information during instantiation, as
            # `self.status()` involves an extra call to the API.
            if api_status == 'VALIDATING':
                self._status = AcQuantumJobStatus.VALIDATING
            elif api_status == 'COMPLETED':
                self._status = AcQuantumJobStatus.DONE
            elif api_status == 'CANCELLED':
                self._status = AcQuantumJobStatus.CANCELLED
                self._cancelled = True
            else:
                self.status()
        self._queue_position = None
        self._is_device = is_device

        def current_utc_time():
            """Gets the current time in UTC format"""
            datetime.datetime.utcnow().replace(tzinfo=datetime.timezone.utc).isoformat()

        self._creation_date = creation_date or current_utc_time()
        self._api_error_msg = None

    def job_id(self):
        self._check_for_submission()
        return self._job_id

    def submit(self, n_qubits=None):
        # type: (int) -> None

        if self._qobj is None:
            raise AcQuantumJobError('Can not find qobj')

        if len(self._qobj.experiments) != 1:
            raise AcQuantumJobError('The qobj must have exactly one (1) experiment.')

        if n_qubits is None:
            n_qubits = self._backend.configuration().n_qubits

        backend_type = self._backend.backend_type()

        if self._job_name is None:
            if self._qobj.experiments[0].header.name:
                self._job_name = self._qobj.experiments[0].header.name
            else:
                self._job_name = self._generate_job_name()
        job_id = None
        try:
            job_id = self._api.create_experiment(n_qubits, backend_type, self._job_name)
            self._job_id = str(job_id)

            gates = self._gates_from_qobj(self._qobj)
            self._api.update_experiment(job_id, gates, code=json.dumps(self._qobj.as_dict()))

            seed = getattr(self._qobj.config, "seed", None)
            self._api.run_experiment(job_id, experiment_type=backend_type, bit_width=n_qubits,
                                     shots=self._qobj.config.shots, seed=seed)
        except AcQuantumRequestError as e:
            if job_id is not None:
                self._api.delete_experiment(job_id)
            raise AcQuantumJobError(e.message)

    def cancel(self):
        # type: () -> bool
        """
        :return: bool: True if job can be cancelled else False.
        :raises: AcQuantumJobError: if there was some unexpected failure in the server
        """
        try:
            # self._api.del(experiment_id=int(self._job_id))
            status = self.status()
            if status is AcQuantumJobStatus.QUEUED:
                result = self._api.get_result(int(self.job_id())).get_results()
                if result:
                    self._api.delete_result(result[-1].result_id)
                    self._status = AcQuantumJobStatus.CANCELLED
                    return True
            else:
                return False
        except AcQuantumRequestError as ex:
            self._status = AcQuantumJobStatus.ERROR
            raise AcQuantumJobError('Error canceling job: {}'.format(ex.message))

    def result(self, timeout=None, wait=5):
        # type: (int, int) -> Result
        """
        Return the result from the job.
        :param timeout: number of seconds to wait for job
        :param wait: time between queries to Alibaba Computing Quantum
        :return: qiskit.Result
        """
        job_response = self._wait_for_result(timeout=timeout, wait=5)
        return self._result_from_job_response(job_response)

    def _wait_for_result(self, timeout=None, wait=5):
        # type: (int, int) -> AcQuantumResultResponse
        self._check_for_submission()
        try:
            job_response = self._wait_for_job(timeout=timeout, wait=wait)
        except AcQuantumRequestError:
            raise AcQuantumJobError('Result query failed')
        status = self.status()
        if status is not AcQuantumJobStatus.DONE:
            raise AcQuantumJobError('Invalid job state. The job should be DONE but it is {}'.format(str(status)))
        return job_response

    def _wait_for_job(self, timeout, wait):
        # type: (int, int) -> AcQuantumResultResponse
        start_time = time.time()
        while self.status() not in JOB_FINAL_STATES:
            elapsed_time = time.time() - start_time
            if timeout is not None and elapsed_time >= timeout:
                raise AcQuantumJobTimeOutError('Timeout while waiting for the job: {}'.format(self._job_id))

            time.sleep(wait)
        if self._cancelled:
            raise AcQuantumJobError('Job result impossible to retrieve. The job was cancelled')

        return self._api.get_result(int(self._job_id))

    def _check_for_submission(self):
        """
        Check if Job was already submitted
        """
        if self._job_id is None:
            raise AcQuantumJobError('You have to submit before asking status or results')

    def status(self):
        # type: () -> AcQuantumJobStatus
        """
        Query the Api to update the status of the job
        :return: The status of the job, once updated
        :raises: AcQuantumJobError: if there was an unknown answer from the server
        """
        if self._status is AcQuantumJobStatus.CANCELLED:
            return self._status

        try:
            result = self._api.get_result(experiment_id=int(self._job_id)).get_results()
            if not result:
                self._status = AcQuantumJobStatus.QUEUED
            else:
                result = result[-1]
        except AcQuantumRequestError as e:
            print(e)
            self._status = AcQuantumJobStatus.ERROR
            return self._status

        if not result.finish_time:
            self._status = AcQuantumJobStatus.RUNNING
            queued, self._queue_position = self._is_job_queued(result)
            if queued:
                self._status = AcQuantumJobStatus.QUEUED
        elif result.exception:
            self._status = AcQuantumJobStatus.ERROR
        elif result.finish_time:
            self._status = AcQuantumJobStatus.DONE
        else:
            raise AcQuantumJobError('Unrecognized answer from server: \n{}'.format(result))
        return self._status

    def _generate_job_name(self):
        # type: () -> str
        return 'Qiskit_generated_{}'.format(self._creation_date)

    def _result_from_job_response(self, job_response):
        # type: (AcQuantumResultResponse) -> Result

        backend = self.backend()  # type: BaseBackend
        config = backend.configuration()  # type: BackendConfiguration
        experiment = self._api.get_experiment(int(self.job_id()))  # type: AcQuantumExperiment

        result_details = {}
        job_results = job_response.get_results()
        if len(job_results) == 1:
            experiment_result = job_results[0]  # type: AcQuantumResult

            counts = dict((hex(int(k, 2)), int(v * experiment_result.shots)) for k, v in experiment_result.data.items())
            self._qobj = Qobj.from_dict(json.loads(experiment.code))
            self._job_name = self._qobj.experiments[0].header.name

            success = experiment_result.exception is None

            result_details = {
                "status": self._status.name,
                "success": success,
                "name": self._job_name,
                "seed": experiment_result.seed,
                "shots": experiment_result.shots,
                "data": {
                    "counts": counts
                },
                "start_time": experiment_result.start_time,
                "finish_time": experiment_result.finish_time,
                "header": self._qobj.experiments[0].header.as_dict()
            }

        from dateutil.parser import parser
        date = parser().parse(result_details['finish_time'])

        result_dict = {
            'results': [result_details],
            'backend_name': config.backend_name,
            'backend_version': config.backend_version,
            'qobj_id': self._qobj.qobj_id,
            'job_id': str(self.job_id()),
            'success': len(job_results) == 1,
            'header': {
                "backend_name": config.backend_name
            },
            "date": date.isoformat()
        }

        result = Result.from_dict(result_dict)

        return result

    @classmethod
    def _is_job_queued(cls, result):
        # type: (AcQuantumResult) -> (bool, int)
        """Checks whether a job has been queued or not."""
        is_queued, position = False, 0
        if not result.finish_time:
            is_queued = True
            position = result.process[1:-1]  # TODO CHECK ATTRIBUTE
        return is_queued, position

    def get_queue_position(self):
        return self._queue_position

    @classmethod
    def _gates_from_qobj(cls, qobj):
        # type: (Qobj) -> List[Gate]
        import qiskit.extensions.standard as standard_gates
        import acquantumconnector.model.gates as ac_gates
        from qiskit.converters import qobj_to_circuits, circuit_to_dag

        qubit_labels = qobj.experiments[0].header.qubit_labels  # type: List[List[Any]]

        qc = qobj_to_circuits(qobj)  # type: QuantumCircuit
        # TODO: do all circuits.
        dag = circuit_to_dag(qc[0])  # type: DAGCircuit

        # Result
        gates = []  # type: List[Gate]

        def get_q_index(qarg):
            qubit, y = qarg
            q_index = qubit_labels.index([qubit.name, y])
            return q_index

        current_layer_x_number = 1
        layer = {}  # type: Dict[str, Union[DAGCircuit, list]]
        for layer in dag.layers():
            layer_dag = layer["graph"]  # type: DAGCircuit
            x_numbers = dict((qubit_labels.index([r.name, i]), current_layer_x_number) for r, i in layer_dag.wires
                             if isinstance(r, QuantumRegister))

            for op_node in layer_dag.get_op_nodes(data=True):
                # Given keys:
                # op_node[1]["cargs"]
                # op_node[1]["condition"]
                # op_node[1]["name"]
                # op_node[1]["qargs"]
                # op_node[1]["type"]
                # op_node[1]["op"]

                if isinstance(op_node[1]["op"], standard_gates.U1Gate):
                    qk_gate = op_node[1]["op"]  # type: standard_gates.U1Gate
                    y = get_q_index(qk_gate.qargs[0])
                    [theta] = qk_gate.param
                    gates.append(ac_gates.RzGate(x_numbers[y], y + y, theta))
                    x_numbers[y] += 1

                elif isinstance(op_node[1]["op"], standard_gates.U2Gate):
                    qk_gate = op_node[1]["op"]  # type: standard_gates.U2Gate
                    y = get_q_index(qk_gate.qargs[0])
                    [phi, lam] = qk_gate.param

                    # ignore global phase alpha!
                    # alpha = lam/2 + phi/2
                    beta = phi
                    delta = lam
                    gamma = math.pi / 2
                    gates.append(ac_gates.RzGate(x_numbers[y], y + 1, delta))
                    x_numbers[y] += 1
                    gates.append(ac_gates.RyGate(x_numbers[y], y + 1, gamma))
                    x_numbers[y] += 1
                    gates.append(ac_gates.RzGate(x_numbers[y], y + 1, beta))
                    x_numbers[y] += 1

                elif isinstance(op_node[1]["op"], standard_gates.U3Gate):
                    qk_gate = op_node[1]["op"]  # type: standard_gates.U3Gate
                    # I assume that all qubits have been transformed into the qubit regsiter 'q'
                    # and that the index simply tells me which wire this is.
                    [theta, phi, lam] = qk_gate.param
                    y = get_q_index(qk_gate.qargs[0])

                    # ignore global phase alpha!
                    # alpha = lam/2 + phi/2
                    beta = phi
                    delta = lam
                    gamma = theta
                    gates.append(ac_gates.RzGate(x_numbers[y], y + 1, delta))
                    x_numbers[y] += 1
                    gates.append(ac_gates.RyGate(x_numbers[y], y + 1, gamma))
                    x_numbers[y] += 1
                    gates.append(ac_gates.RzGate(x_numbers[y], y + 1, beta))
                    x_numbers[y] += 1

                elif isinstance(op_node[1]["op"], standard_gates.HGate):
                    qk_gate = op_node[1]["op"]  # type: standard_gates.HGate
                    y = get_q_index(qk_gate.qargs[0])
                    gates.append(ac_gates.HGate(x_numbers[y], y + 1))
                    x_numbers[y] += 1

                elif isinstance(op_node[1]["op"], standard_gates.RXGate):
                    qk_gate = op_node[1]["op"]  # type: standard_gates.RXGate
                    y = get_q_index(qk_gate.qargs[0])
                    [theta] = qk_gate.param
                    gates.append(ac_gates.RxGate(x_numbers[y], y + 1, theta))
                    x_numbers[y] += 1

                elif isinstance(op_node[1]["op"], standard_gates.RYGate):
                    qk_gate = op_node[1]["op"]  # type: standard_gates.RYGate
                    y = get_q_index(qk_gate.qargs[0])
                    [theta] = qk_gate.param
                    gates.append(ac_gates.RyGate(x_numbers[y], y + 1, theta))
                    x_numbers[y] += 1

                elif isinstance(op_node[1]["op"], standard_gates.RZGate):
                    qk_gate = op_node[1]["op"]  # type: standard_gates.RZGate
                    y = get_q_index(qk_gate.qargs[0])
                    [theta] = qk_gate.param
                    gates.append(ac_gates.RzGate(x_numbers[y], y + 1, theta))
                    x_numbers[y] += 1

                elif isinstance(op_node[1]["op"], standard_gates.XGate):
                    qk_gate = op_node[1]["op"]  # type: standard_gates.XGate
                    y = get_q_index(qk_gate.qargs[0])
                    gates.append(ac_gates.XGate(x_numbers[y], y + 1))
                    x_numbers[y] += 1

                elif isinstance(op_node[1]["op"], standard_gates.YGate):
                    qk_gate = op_node[1]["op"]  # type: standard_gates.YGate
                    y = get_q_index(qk_gate.qargs[0])
                    gates.append(ac_gates.YGate(x_numbers[y], y + 1))
                    x_numbers[y] += 1

                elif isinstance(op_node[1]["op"], standard_gates.ZGate):
                    qk_gate = op_node[1]["op"]  # type: standard_gates.ZGate
                    y = get_q_index(qk_gate.qargs[0])
                    gates.append(ac_gates.ZGate(x_numbers[y], y + 1))
                    x_numbers[y] += 1

                elif isinstance(op_node[1]["op"], standard_gates.SGate):
                    qk_gate = op_node[1]["op"]  # type: standard_gates.SGate
                    y = get_q_index(qk_gate.qargs[0])
                    gates.append(ac_gates.SGate(x_numbers[y], y + 1))
                    x_numbers[y] += 1

                elif isinstance(op_node[1]["op"], standard_gates.SdgGate):
                    qk_gate = op_node[1]["op"]  # type: standard_gates.SdgGate
                    y = get_q_index(qk_gate.qargs[0])
                    gates.append(ac_gates.SDag(x_numbers[y], y + 1))
                    x_numbers[y] += 1

                elif isinstance(op_node[1]["op"], standard_gates.TGate):
                    qk_gate = op_node[1]["op"]  # type: standard_gates.TGate
                    y = get_q_index(qk_gate.qargs[0])
                    gates.append(ac_gates.TGate(x_numbers[y], y + 1))
                    x_numbers[y] += 1

                elif isinstance(op_node[1]["op"], standard_gates.TdgGate):
                    qk_gate = op_node[1]["op"]  # type: standard_gates.TdgGate
                    y = get_q_index(qk_gate.qargs[0])
                    gates.append(ac_gates.TDag(x_numbers[y], y + 1))
                    x_numbers[y] += 1

                elif isinstance(op_node[1]["op"], standard_gates.CrzGate):
                    qk_gate = op_node[1]["op"]  # type: standard_gates.CrzGate
                    y = get_q_index(qk_gate.qargs[0])
                    gates.append(ac_gates.CPhase(x_numbers[y], y + 1))
                    x_numbers[y] += 1

                elif isinstance(op_node[1]["op"], standard_gates.CnotGate):
                    cx = op_node[1]["op"]  # type: standard_gates.CnotGate
                    y1 = get_q_index(cx.qargs[0])
                    y2 = get_q_index(cx.qargs[1])
                    gates.append(ac_gates.HGate(x_numbers[y2], y2 + 1))
                    x_numbers[y1] += 1
                    x_numbers[y2] += 1
                    gates.append(ac_gates.CPhase(x_numbers[y2], [y1 + 1, y2 + 1]))
                    x_numbers[y1] += 1
                    x_numbers[y2] += 1
                    gates.append(ac_gates.HGate(x_numbers[y2], y2 + 1))
                    x_numbers[y1] += 1
                    x_numbers[y2] += 1

                elif isinstance(op_node[1]["op"], standard_gates.ToffoliGate):
                    cx = op_node[1]["op"]  # type: standard_gates.ToffoliGate
                    y1 = get_q_index(cx.qargs[0])
                    y2 = get_q_index(cx.qargs[1])
                    y3 = get_q_index(cx.qargs[2])
                    gates.append(ac_gates.HGate(x_numbers[y3], y3 + 1))
                    x_numbers[y1] += 1
                    x_numbers[y2] += 1
                    x_numbers[y3] += 1
                    gates.append(ac_gates.CCPhase(x_numbers[y3], [y1 + 1, y2 + 1, y3 + 1]))
                    x_numbers[y1] += 1
                    x_numbers[y2] += 1
                    x_numbers[y3] += 1
                    gates.append(ac_gates.HGate(x_numbers[y3], y3 + 1))
                    x_numbers[y1] += 1
                    x_numbers[y2] += 1
                    x_numbers[y3] += 1

                elif isinstance(op_node[1]["op"], qiskit.circuit.Measure):
                    measure = op_node[1]["op"]  # type: qiskit.circuit.Measure
                    y = get_q_index(measure.qargs[0])
                    gates.append(ac_gates.Measure(x_numbers[y], y + 1))
                    x_numbers[y] += 1

            current_layer_x_number = max(x_numbers.values())

        return gates



#  Copyright (c) 2019.  Carsten Blank
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from setuptools import setup

with open("providers/_version.py") as f:
    version = f.readlines()[-1].split()[-1].strip("\"'")

with open("README.rst", "r") as fh:
    long_description = fh.read()

with open("requirements.txt") as f:
    requirements = f.readlines()

info = {
    'name': 'acquantum-qiskit',
    'version': version,
    'author': 'Carsten Blank',
    'author_email': 'blank@data-cybernetics.com',
    'description': 'A qiskit provider for the Alibaba\'s quantum computer.',
    'long_description': long_description,
    'url': 'https://github.com/sebboer/acquantum_qiskit',
    'install_requires': requirements,
    'packages': [
        'providers',
        'providers.acquantum',
        'providers.acquantum.credentials',
        'providers.acquantum.models'
    ],
    'license': 'Apache 2.0',
}

classifiers = [
    'Development Status :: 4 - Beta',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3 :: Only',
    'License :: OSI Approved :: Apache Software License',
    'Operating System :: OS Independent',
]

setup(classifiers=classifiers, **info)



#  Copyright (c) 2019.  Carsten Blank
# 
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
# 
#       http://www.apache.org/licenses/LICENSE-2.0
# 
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import warnings
from collections import OrderedDict

from qiskit.providers import BaseProvider

from acquantumconnector.credentials.credentials import AcQuantumCredentials
from .acquantumbackend import AcQuantumBackend
from .acquantumerrors import AcQuantumAccountError, AcQuantumBackendError
from .acquantumsingleprovider import AcQuantumSingleProvider
from .credentials import discover_credentials


class AcQuantumProvider(BaseProvider):

    def __init__(self):
        super().__init__()

        self._accounts = OrderedDict()

    def backends(self, name=None, **kwargs):
        # type: (str, dict) -> [AcQuantumBackend]
        """
        :param name: name of the backend
        :param kwargs: kwargs for filtering not yet implemented
        :return:
        """

        providers = [provider for provider in self._accounts.values()]

        # Aggregate the list of filtered backends.
        backends = []
        for provider in providers:
            backends = backends + list(provider.backends(
                name=name, **kwargs))

        if not backends:
            raise AcQuantumBackendError('zero backends found')
        return backends

    def load_account(self, credentials=None):
        if credentials is None:
            self._append_account(discover_credentials())
        else:
            self._append_account(credentials)

        if not self._accounts:
            raise AcQuantumAccountError('No AcQuantum credentials found.')

    def _append_account(self, credentials: AcQuantumCredentials):
        if credentials.user_name in self._accounts.keys():
            warnings.warn('Credentials are already in use.')

        single_provider = AcQuantumSingleProvider(credentials, self)
        self._accounts[credentials.user_name] = single_provider

    @staticmethod
    def _aliased_backend_names(self):
        pass

    @staticmethod
    def _deprecated_backend_names(self):
        pass
